#include "stdafx.h"
#include <memory.h>
#include "AvSyncMultiPlaybackWrapper.h"


// Function     : CAvSyncMultiPlaybackWrapper::CAvSyncMultiPlaybackWrapper
// Description  : Constructor
// Parameter    : 
//              | [in] hDlg             @Dialog window handle
//              | [in] ahDisplay        @Array of multiple display window handle
//              | [in] dwChannelNum     @Specified number of channels
//              | [in] hBMP             @Bitmap handle
//              | [in] pfStatus         @Status callback function
//              | [in] pfDisplay        @Display callback function
//              | [in] dwStatusContext  @Status context
//              | [in] dwDisplayContext @Display context
// Return value : N/A
//
CAvSyncMultiPlaybackWrapper::CAvSyncMultiPlaybackWrapper(HWND hDlg, HWND ahDisplay[], DWORD dwChannelNum, HBITMAP hBMP, LPSTATUSCALLBACK pfStatus, LPDISPLAYCALLBACK pfDisplay, DWORD dwStatusContext, DWORD dwDisplayContext)
{
	m_hAvSync = NULL;

	m_dwChannelNum = dwChannelNum;

	m_phPlaybackChannel = new HANDLE [m_dwChannelNum];  // Initialize multiple playback channel handles
	
	memset(&m_tParam, 0, sizeof(m_tParam));

	m_tParam.hDlg     = hDlg;
	m_tParam.hBMP     = hBMP;
	
	m_tParam.pfStatus  = pfStatus;
	m_tParam.pfDisplay = pfDisplay;
	
	m_tParam.dwStatusContext  = dwStatusContext;
	m_tParam.dwDisplayContext = dwDisplayContext;

	m_tParam.phDisplay = new HWND [m_dwChannelNum];

	memcpy_s(m_tParam.phDisplay, m_dwChannelNum * sizeof(HWND), ahDisplay, m_dwChannelNum * sizeof(HWND));  // Get multiple display window handles

	InitialModule();  // Initialize AvSynchronizer and create playback channels
}

// Function     : CAvSyncMultiPlaybackWrapper::~CAvSyncMultiPlaybackWrapper
// Description  : Destructor
// Parameter    : N/A
// Return value : N/A
//
CAvSyncMultiPlaybackWrapper::~CAvSyncMultiPlaybackWrapper()
{
	ReleaseModule();  // Delete playback channels and release AvSynchronizer

	delete [] m_phPlaybackChannel;

	DWORD dwCHIndex;

	for	(dwCHIndex = 0; dwCHIndex < m_dwChannelNum; ++dwCHIndex)
	{
		::DeleteObject(m_tParam.phDisplay[dwCHIndex]);  // Destroy the copy of display window handles from dialog
	}
	
	delete [] m_tParam.phDisplay; 
}

// Function     : CAvSyncMultiPlaybackWrapper::InitialModule
// Description  : Initialize AvSynchronizer and create playback channels
// Parameter    : N/A
// Return value :
//              | S_OK   @Success
//              | S_FAIL @Fail
//
SCODE CAvSyncMultiPlaybackWrapper::InitialModule()
{	
	SCODE scRet;

	// STEP 1: Initialize AvSynchronizer

	// Notice!
	// Here we want to print out the channel index on the bottom-right of display window, hence the flag "CAPTION_ON_GRAPH" is set.
	// Setting the flag "CAPTION_ON_GRAPH" can cause lower performance due to a large number of callback times
	// User can just pass 0 to the "dwflag" field if the DC of whole display window is not used in AP.
	//
	scRet = AvSynchronizer_InitialEx(&m_hAvSync, m_tParam.pfStatus, m_tParam.pfDisplay, m_tParam.hDlg, AUDIOOUT_FOCUS_NORMAL, 0, AVSYNCHRONIZER_VERSION, CAPTION_ON_GRAPH, RGB(0, 0, 0));

	if	(IS_FAIL(scRet))
	{
		sprintf_s(m_szErrMsgBuf, sizeof(m_szErrMsgBuf), "AvSynchronizer_InitialEx fail! (ERROR: %X)", scRet);

		MessageBox(NULL, m_szErrMsgBuf, NULL, MB_OK);

		return S_FAIL;
	}

	// STEP 2: Create multiple playback channels
	if	(m_hAvSync != NULL)
	{
		DWORD dwCHIndex;

		for	(dwCHIndex = 0; dwCHIndex < m_dwChannelNum; ++dwCHIndex)
		{
			scRet = CreateChannel(dwCHIndex);

			if	(scRet != S_OK)
			{
				return S_FAIL;
			}
		}
	}

	return S_OK;
}

// Function     : CAvSyncMultiPlaybackWrapper::ReleaseModule
// Description  : Delete playback channels and release AvSynchronizer
// Parameter    : N/A
// Return value :
//              | S_OK   @Success
//              | S_FAIL @Fail
//
SCODE CAvSyncMultiPlaybackWrapper::ReleaseModule()
{
	SCODE scRet;
	
	if	(m_hAvSync != NULL)
	{
		DWORD dwCHIndex;
			
		// STEP 1: Delete multiple playback channels
		for	(dwCHIndex = 0; dwCHIndex < m_dwChannelNum; ++dwCHIndex)
		{
			if	(m_phPlaybackChannel[dwCHIndex] != NULL)
			{			
				scRet = DeleteChannel(dwCHIndex);

				if	(scRet != S_OK)
				{
					return S_FAIL;
				}
			}
		}

		// STEP 2: Release AvSynchronizer
		scRet = AvSynchronizer_Release(&m_hAvSync);

		if	(IS_FAIL(scRet))
		{
			sprintf_s(m_szErrMsgBuf, sizeof(m_szErrMsgBuf), "AvSynchronizer_Release fail! (ERROR: %X)", scRet);

			MessageBox(NULL, m_szErrMsgBuf, NULL, MB_OK);

			return S_FAIL;
		}
	}

	return S_OK;
}

// Function     : CAvSyncMultiPlaybackWrapper::CreateChannel
// Description  : Create playback channel
// Parameter    : 
//              | [in] dwCHIndex @Channel index
// Return value :
//              | S_OK   @Success
//              | S_FAIL @Fail
//
SCODE CAvSyncMultiPlaybackWrapper::CreateChannel(DWORD dwCHIndex)
{	
	// STEP 1: Set channel options
	TPBCHOPTION tPBCHOption;

	memset(&tPBCHOption, 0, sizeof(tPBCHOption));

	tPBCHOption.dwFlags = PBCH_STATUSCB | PBCH_DISPLAYCB | PBCH_STATUSCONTEXT | PBCH_DISPLAYCONTEXT | PBCH_ALLBORDER | PBCH_BITMAP;

	tPBCHOption.pfStatus  = m_tParam.pfStatus;
	tPBCHOption.pfDisplay = m_tParam.pfDisplay; 

	tPBCHOption.dwStatusContext  = m_tParam.dwStatusContext;
	tPBCHOption.dwDisplayContext = m_tParam.dwDisplayContext + dwCHIndex;  // Set the display context as the channel index
	
	tPBCHOption.dwTopBorderSize    = 20;
	tPBCHOption.dwLeftBorderSize   = 0;
	tPBCHOption.dwBottomBorderSize = 0;
	tPBCHOption.dwRightBorderSize  = 0;

	tPBCHOption.hBMP     = m_tParam.hBMP;
	tPBCHOption.hDisplay = m_tParam.phDisplay[dwCHIndex];
	
	SCODE scRet;

	// STEP 2: Create playback channel
	scRet = AvSynchronizer_CreatePlaybackChannel(m_hAvSync, &m_phPlaybackChannel[dwCHIndex], tPBCHOption);

	if	(IS_FAIL(scRet))
	{
		sprintf_s(m_szErrMsgBuf, sizeof(m_szErrMsgBuf), "AvSynchronizer_CreatePlaybackChannel fail! (ERROR: %X)", scRet);

		OutputDebugString(m_szErrMsgBuf);

		return S_FAIL;
	}
	
	return S_OK;
}

// Function     : CAvSyncMultiPlaybackWrapper::DeleteChannel
// Description  : Delete playback channel
// Parameter    : 
//              | [in] dwCHIndex @Channel index
// Return value :
//              | S_OK   @Success
//              | S_FAIL @Fail
//
SCODE CAvSyncMultiPlaybackWrapper::DeleteChannel(DWORD dwCHIndex)
{
	SCODE scRet;

	// Delete playback channel 
	scRet = AvSynchronizer_DeleteChannel(m_hAvSync, &m_phPlaybackChannel[dwCHIndex]);

	if	(IS_FAIL(scRet))
	{
		sprintf_s(m_szErrMsgBuf, sizeof(m_szErrMsgBuf), "AvSynchronizer_DeleteChannel fail! (ERROR: %X)", scRet);

		OutputDebugString(m_szErrMsgBuf);

		return S_FAIL;
	}

	return S_OK;
}

// Function     : CAvSyncMultiPlaybackWrapper::Play
// Description  : Start playback channel
// Parameter    : 
//              | [in] dwCHIndex @Channel index
// Return value :
//              | S_OK   @Success
//              | S_FAIL @Fail
//
SCODE CAvSyncMultiPlaybackWrapper::Play(DWORD dwCHIndex)
{
	SCODE scRet;
	
	// Start playback channel	
	if	(m_phPlaybackChannel[dwCHIndex] != NULL)
	{		
		scRet = AvSynchronizer_StartChannel(m_phPlaybackChannel[dwCHIndex], 0);

		if	(IS_FAIL(scRet))
		{
			sprintf_s(m_szErrMsgBuf, sizeof(m_szErrMsgBuf), "AvSynchronizer_StartChannel fail! (ERROR: %X)", scRet);

			OutputDebugString(m_szErrMsgBuf);

			return S_FAIL;
		}
	}
	
	return S_OK;
}

// Function     : CAvSyncMultiPlaybackWrapper::Stop
// Description  : Stop playback channel
// Parameter    :
//              | [in] dwCHIndex @Channel index
// Return value :
//              | S_OK   @Success
//              | S_FAIL @Fail
//
SCODE CAvSyncMultiPlaybackWrapper::Stop(DWORD dwCHIndex)
{	
	SCODE scRet;
	
	// Stop playback channel
	if	(m_phPlaybackChannel[dwCHIndex] != NULL)
	{
		scRet = AvSynchronizer_StopChannel(m_phPlaybackChannel[dwCHIndex]);

		if	(IS_FAIL(scRet))
		{
			sprintf_s(m_szErrMsgBuf, sizeof(m_szErrMsgBuf), "AvSynchronizer_StopChannel fail! (ERROR: %X)", scRet);

			OutputDebugString(m_szErrMsgBuf);

			return S_FAIL;
		}

		// Show bitmap and reset play speed to normal
		TUPDATEPBCHANNELOPTION tUpdatePBCHOption;

		memset(&tUpdatePBCHOption, 0, sizeof(tUpdatePBCHOption));

		DWORD dwNormalPlaySpeed = 1;

		tUpdatePBCHOption.dwFlags = UPCH_SHOW_BMP | UPCH_SETSPEED;
		tUpdatePBCHOption.dwFast  = dwNormalPlaySpeed;
		tUpdatePBCHOption.dwSlow  = dwNormalPlaySpeed;
			
		scRet = AvSynchronizer_UpdatePlaybackChannelSettings(m_phPlaybackChannel[dwCHIndex], tUpdatePBCHOption);

		if	(IS_FAIL(scRet))
		{
			sprintf_s(m_szErrMsgBuf, sizeof(m_szErrMsgBuf), "AvSynchronizer_UpdatePlaybackChannelSettings fail! (ERROR: %X)\n", scRet);

			OutputDebugString(m_szErrMsgBuf);

			return S_FAIL;
		}
	}

	return S_OK;
}

// Function     : CAvSyncMultiPlaybackWrapper::Pause
// Description  : Pause or Resume playback channel
// Parameter    :
//              | [in] dwCHIndex @Channel index
//              | [in] bPause    @TRUE : pause, FALSE : resume
// Return value :
//              | S_OK   @Success
//              | S_FAIL @Fail
//
SCODE CAvSyncMultiPlaybackWrapper::Pause(DWORD dwCHIndex, BOOL bPause)
{
	TUPDATEPBCHANNELOPTION tUpdatePBCHOption;

	memset(&tUpdatePBCHOption, 0, sizeof(tUpdatePBCHOption));

	tUpdatePBCHOption.dwFlags = UPCH_PAUSE;
	tUpdatePBCHOption.bPause  = bPause;

	SCODE scRet;
			
	if	(m_phPlaybackChannel[dwCHIndex] != NULL)
	{	
		scRet = AvSynchronizer_UpdatePlaybackChannelSettings(m_phPlaybackChannel[dwCHIndex], tUpdatePBCHOption);

		if	(IS_FAIL(scRet))
		{
			sprintf_s(m_szErrMsgBuf, sizeof(m_szErrMsgBuf), "AvSynchronizer_UpdatePlaybackChannelSettings fail! (ERROR: %X)\n", scRet);

			OutputDebugString(m_szErrMsgBuf);

			return S_FAIL;
		}
	}

	return S_OK;
}

// Function     : CAvSyncMultiPlaybackWrapper::Forward
// Description  : Change play speed of playback channel
// Parameter    :
//              | [in] dwCHIndex @Channel index
//              | [in] dwFast    @Multiple of normal play speed
//              | [in] dwSlow    @Divisor of normal play speed
// Return value :
//              | S_OK   @Success
//              | S_FAIL @Fail
//
SCODE CAvSyncMultiPlaybackWrapper::Forward(DWORD dwCHIndex, DWORD dwFast, DWORD dwSlow)
{
	TUPDATEPBCHANNELOPTION tUpdatePBCHOption;

	memset(&tUpdatePBCHOption, 0, sizeof(tUpdatePBCHOption));

	tUpdatePBCHOption.dwFlags = UPCH_SETSPEED;
	tUpdatePBCHOption.dwFast  = dwFast;
	tUpdatePBCHOption.dwSlow  = dwSlow;

	SCODE scRet;

	if	(m_phPlaybackChannel[dwCHIndex] != NULL)
	{			
		scRet = AvSynchronizer_UpdatePlaybackChannelSettings(m_phPlaybackChannel[dwCHIndex], tUpdatePBCHOption);

		if	(IS_FAIL(scRet))
		{
			sprintf_s(m_szErrMsgBuf, sizeof(m_szErrMsgBuf), "AvSynchronizer_UpdatePlaybackChannelSettings fail! (ERROR: %X)\n", scRet);

			OutputDebugString(m_szErrMsgBuf);

			return S_FAIL;
		}
	}

	return S_OK;
}

// Function     : CAvSyncMultiPlaybackWrapper::NextFrame
// Description  : Jump to next frame. The playback channel must be paused in advance!
// Parameter    :
//              | [in] dwCHIndex @Channel index
// Return value :
//              | S_OK   @Success
//              | S_FAIL @Fail
//
SCODE CAvSyncMultiPlaybackWrapper::NextFrame(DWORD dwCHIndex)
{
	TUPDATEPBCHANNELOPTION tUpdatePBCHOption;

	memset(&tUpdatePBCHOption, 0, sizeof(tUpdatePBCHOption));

	tUpdatePBCHOption.dwFlags = PBCH_NEXTFRAME;

	SCODE scRet;
			
	if	(m_phPlaybackChannel[dwCHIndex] != NULL)
	{			
		scRet = AvSynchronizer_UpdatePlaybackChannelSettings(m_phPlaybackChannel[dwCHIndex], tUpdatePBCHOption);

		if	(IS_FAIL(scRet))
		{
			sprintf_s(m_szErrMsgBuf, sizeof(m_szErrMsgBuf), "AvSynchronizer_UpdatePlaybackChannelSettings fail! (ERROR: %X)\n", scRet);

			OutputDebugString(m_szErrMsgBuf);

			return S_FAIL;
		}
	}

	return S_OK;
}

// Function     : CAvSyncMultiPlaybackWrapper::UpdatePlaybackChannelMediaType
// Description  : Update channel media type
// Parameter    :
//              | [in] dwCHIndex   @Channel index
//              | [in] dwMediaType @Media type
// Return value :
//              | S_OK   @Success
//              | S_FAIL @Fail
//
SCODE CAvSyncMultiPlaybackWrapper::UpdatePlaybackChannelMediaType(DWORD dwCHIndex, DWORD dwMediaType)
{
	TUPDATEPBCHANNELOPTION tUpdatePBCHOption;

	memset(&tUpdatePBCHOption, 0, sizeof(TUPDATEPBCHANNELOPTION));

	tUpdatePBCHOption.dwFlags = UPCH_MEDIATYPE;
	tUpdatePBCHOption.tMediaType = EMEDIATYPE(dwMediaType);

	SCODE scRet;

	if	(m_phPlaybackChannel[dwCHIndex] != NULL)
	{
		scRet = AvSynchronizer_UpdatePlaybackChannelSettings(m_phPlaybackChannel[dwCHIndex], tUpdatePBCHOption);

		if	(IS_FAIL(scRet))
		{
			sprintf_s(m_szErrMsgBuf, sizeof(m_szErrMsgBuf), "AvSynchronizer_UpdatePlaybackChannelSettings fail! (ERROR: %X)\n", scRet);

			OutputDebugString(m_szErrMsgBuf);

			return S_FAIL;
		}
	}

	return S_OK;
}

// Function     : CAvSyncMultiPlaybackWrapper::InputDataPacketToChannel
// Description  : Input the parsed data packet info to playback channel
// Parameter    :
//              | [in] dwCHIndex    @Channel index
//              | [in] ptPacketInfoV3 @Input media data packet info
// Return value :
//              | S_OK   @Success
//              | S_FAIL @Fail
//
SCODE CAvSyncMultiPlaybackWrapper::InputDataPacketToChannel(DWORD dwCHIndex, TMediaDataPacketInfoV3 *ptPacketInfoV3)
{
	// Notice!
	// If the internal buffer of "PLAYBACK CHANNEL" is full, it will block the calling thread by default. 
	// User can call AvSynchronizer_SetChannelOption with SETCH_NOT_BLOCK_WHEN_QUEUE_FULL to enable or disable this function.
	
	SCODE scRet;
	
	if	(m_phPlaybackChannel[dwCHIndex] != NULL)
	{
		TMediaDataPacketInfo *ptPacketInfo = (TMediaDataPacketInfo *)ptPacketInfoV3;  // Cast "ptPacketInfoV3" to (TMediaDataPacketInfo *)

		// Input the parsed data packet info to playback channel
		scRet = AvSynchronizer_InputPlaybackMediaFrame(m_phPlaybackChannel[dwCHIndex], ptPacketInfo);

		if	(IS_FAIL(scRet))
		{
			sprintf_s(m_szErrMsgBuf, sizeof(m_szErrMsgBuf), "AvSynchronizer_InputPlaybackMediaFrame fail! (ERROR: %X)\n", scRet);

			OutputDebugString(m_szErrMsgBuf);

			return S_FAIL;
		}
	}

	return S_OK;
}

// Function     : CAvSyncMultiPlaybackWrapper::WaitAVQueueEmpty
// Description  : Wait until the audio and video elements in AvSynchronizer's queue are all played 
// Parameter    :
//              | [in] dwCHIndex @Channel index
// Return value :
//              | S_OK   @Success
//              | S_FAIL @Fail
//
SCODE CAvSyncMultiPlaybackWrapper::WaitAVQueueEmpty(DWORD dwCHIndex)
{
	TAVSYNCQUEUESTATUS tAvSyncQueueStatus;

	SCODE scRet;
	
	if	(m_phPlaybackChannel[dwCHIndex] != NULL)
	{			
		// Wait until all audio and video elements are played.
		// When all elements are played, the audio buffer queue is unused now, hence the dwAudioQueueElements would be equal to dwAudioQueueSize.
		// On the other hand, the video buffer will keep one unit to redraw the last frame, hence the dwVideoQueueElements = dwVideoQueueSize - 1.
		//
		do{
			Sleep(20);

			memset(&tAvSyncQueueStatus, 0, sizeof(tAvSyncQueueStatus));

			// Get remaining audio and video elements in queue
			scRet = AvSynchronizer_GetRemainingQueueElementCount(m_phPlaybackChannel[dwCHIndex], &tAvSyncQueueStatus);

			if	(IS_FAIL(scRet))
			{
				sprintf_s(m_szErrMsgBuf, sizeof(m_szErrMsgBuf), "AvSynchronizer_GetRemainingQueueElementCount fail! (ERROR: %X)\n", scRet);

				OutputDebugString(m_szErrMsgBuf);

				return S_FAIL;
			}
		
		}while	((tAvSyncQueueStatus.dwAudioQueueElements != tAvSyncQueueStatus.dwAudioQueueSize) || (tAvSyncQueueStatus.dwVideoQueueElements != tAvSyncQueueStatus.dwVideoQueueSize - 1));
	}
	
	return S_OK;
}
